package java.lang.ref;

import sun.misc.SharedSecrets;
import sun.misc.VM;
import java.security.AccessController;
import java.security.PrivilegedAction;
import sun.misc.JavaLangAccess;

final class Finalizer extends FinalReference<Object>
{
    private static ReferenceQueue<Object> queue;
    private static Finalizer unfinalized;
    private static final Object lock;
    private Finalizer next;
    private Finalizer prev;
    
    private boolean hasBeenFinalized() {
        return this.next == this;
    }
    
    private void add() {
        synchronized (Finalizer.lock) {
            if (Finalizer.unfinalized != null) {
                this.next = Finalizer.unfinalized;
                Finalizer.unfinalized.prev = this;
            }
            Finalizer.unfinalized = this;
        }
    }
    
    private void remove() {
        synchronized (Finalizer.lock) {
            if (Finalizer.unfinalized == this) {
                if (this.next != null) {
                    Finalizer.unfinalized = this.next;
                }
                else {
                    Finalizer.unfinalized = this.prev;
                }
            }
            if (this.next != null) {
                this.next.prev = this.prev;
            }
            if (this.prev != null) {
                this.prev.next = this.next;
            }
            this.next = this;
            this.prev = this;
        }
    }
    
    private Finalizer(Object o) {
        super(o, Finalizer.queue);
        this.next = null;
        this.prev = null;
        this.add();
    }
    
    static ReferenceQueue<Object> getQueue() {
        return Finalizer.queue;
    }
    
    static void register(Object o) {
        new Finalizer(o);
    }
    
    private void runFinalizer(JavaLangAccess javaLangAccess) {
        synchronized (this) {
            if (this.hasBeenFinalized()) {
                return;
            }
            this.remove();
        }
        try {
            Object value = this.get();
            if (value != null && !(value instanceof Enum)) {
                javaLangAccess.invokeFinalize(value);
            }
        }
        catch (Throwable t) {}
        super.clear();
    }
    
    private static void forkSecondaryFinalizer(final Runnable runnable) {
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                ThreadGroup group;
                for (group = (group = Thread.currentThread().getThreadGroup()); group != null; , group = group.getParent()) {}
                Thread o = new Thread(group, runnable, "Secondary finalizer");
                o.start();
                try {
                    o.join();
                }
                catch (InterruptedException ex) {
                    Thread.currentThread().interrupt();
                }
                return null;
            }
        });
    }
    
    static void runFinalization() {
        if (!VM.isBooted()) {
            return;
        }
        forkSecondaryFinalizer(new Runnable() {
            private volatile boolean running;
            
            @Override
            public void run() {
                if (this.running) {
                    return;
                }
                JavaLangAccess javaLangAccess = SharedSecrets.getJavaLangAccess();
                this.running = true;
                while (true) {
                    Finalizer finalizer = (Finalizer)Finalizer.queue.poll();
                    if (finalizer == null) {
                        break;
                    }
                    Finalizer.access$100(finalizer, javaLangAccess);
                }
                return;
            }
        });
        return;
    }
    
    static void runAllFinalizers() {
        if (!VM.isBooted()) {
            return;
        }
        forkSecondaryFinalizer(new Runnable() {
            private volatile boolean running;
            
            @Override
            public void run() {
                if (this.running) {
                    return;
                }
                JavaLangAccess javaLangAccess = SharedSecrets.getJavaLangAccess();
                this.running = true;
                while (true) {
                    Finalizer access$300;
                    synchronized (Finalizer.lock) {
                        access$300 = Finalizer.unfinalized;
                        if (access$300 == null) {
                            break;
                        }
                        Finalizer.unfinalized = access$300.next;
                    }
                    Finalizer.access$100(access$300, javaLangAccess);
                }
                return;
            }
        });
        return;
    }
    
    static void access$100(Finalizer finalizer, JavaLangAccess javaLangAccess) {
        finalizer.runFinalizer(javaLangAccess);
    }
    
    static {
        Finalizer.queue = new ReferenceQueue();
        Finalizer.unfinalized = null;
        lock = new Object();
        ThreadGroup threadGroup;
        for (threadGroup = (threadGroup = Thread.currentThread().getThreadGroup()); threadGroup != null; , threadGroup = threadGroup.getParent()) {}
        Finalizer.FinalizerThread o = new FinalizerThread(threadGroup);
        o.setPriority(8);
        o.setDaemon(true);
        o.start();
    }
    
    private static class FinalizerThread extends Thread
    {
        private volatile boolean running;
        
        FinalizerThread(ThreadGroup group) {
            super(group, "Finalizer");
        }
        
        @Override
        public void run() {
            if (this.running) {
                return;
            }
            while (!VM.isBooted()) {
                try {
                    VM.awaitBooted();
                }
                catch (InterruptedException ex) {}
            }
            JavaLangAccess javaLangAccess = SharedSecrets.getJavaLangAccess();
            this.running = true;
            while (true) {
                while (true) {
                    try {
                        Finalizer.access$100((Finalizer)Finalizer.queue.remove(), javaLangAccess);
                        continue;
                    }
                    catch (InterruptedException ex2) {}
                }
            }
        }
    }
}
